"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Tracker = function () {
  function Tracker() {
    _classCallCheck(this, Tracker);
  }

  _createClass(Tracker, null, [{
    key: "flush",


    /**
    * @summary A Computation object represents code that is repeatedly rerun
    * in response to
    * reactive data changes. Computations don't have return values; they just
    * perform actions, such as rerendering a template on the screen. Computations
    * are created using Tracker.autorun. Use stop to prevent further rerunning of a
    * computation.
    * @instancename computation
    */
    value: function flush(options) {
      Tracker._runFlush({ finishSynchronously: true,
        throwFirstError: options && options._throwFirstError });
    }
  }, {
    key: "_runFlush",
    value: function _runFlush(options) {
      // XXX What part of the comment below is still true? (We no longer
      // have Spark)
      //
      // Nested flush could plausibly happen if, say, a flush causes
      // DOM mutation, which causes a "blur" event, which runs an
      // app event handler that calls Tracker.flush.  At the moment
      // Spark blocks event handlers during DOM mutation anyway,
      // because the LiveRange tree isn't valid.  And we don't have
      // any useful notion of a nested flush.
      //
      // https://app.asana.com/0/159908330244/385138233856
      if (inFlush) throw new Error("Can't call Tracker.flush while flushing");

      if (inCompute) throw new Error("Can't flush inside Tracker.autorun");

      options = options || {};

      inFlush = true;
      willFlush = true;
      throwFirstError = !!options.throwFirstError;

      var recomputedCount = 0;
      var finishedTry = false;
      try {
        while (pendingComputations.length || afterFlushCallbacks.length) {

          // recompute all pending computations
          while (pendingComputations.length) {
            var comp = pendingComputations.shift();
            comp._recompute();
            if (comp._needsRecompute()) {
              pendingComputations.unshift(comp);
            }

            if (!options.finishSynchronously && ++recomputedCount > 1000) {
              finishedTry = true;
              return;
            }
          }

          if (afterFlushCallbacks.length) {
            // call one afterFlush callback, which may
            // invalidate more computations
            var func = afterFlushCallbacks.shift();
            try {
              func();
            } catch (e) {
              _throwOrLog("afterFlush", e);
            }
          }
        }
        finishedTry = true;
      } finally {
        if (!finishedTry) {
          // we're erroring due to throwFirstError being true.
          inFlush = false; // needed before calling `Tracker.flush()` again
          // finish flushing
          Tracker._runFlush({
            finishSynchronously: options.finishSynchronously,
            throwFirstError: false
          });
        }
        willFlush = false;
        inFlush = false;
        if (pendingComputations.length || afterFlushCallbacks.length) {
          // We're yielding because we ran a bunch of computations and we aren't
          // required to finish synchronously, so we'd like to give the event loop a
          // chance. We should flush again soon.
          if (options.finishSynchronously) {
            throw new Error("still have more to do?"); // shouldn't happen
          }
          setTimeout(requireFlush, 10);
        }
      }
    }
  }, {
    key: "autorun",
    value: function autorun(f, options) {
      if (typeof f !== 'function') throw new Error('Tracker.autorun requires a function argument');

      options = options || {};

      constructingComputation = true;
      var c = new Tracker.Computation(f, Tracker.currentComputation, options.onError);

      if (Tracker.active) Tracker.onInvalidate(function () {
        c.stop();
      });

      return c;
    }
  }, {
    key: "nonreactive",
    value: function nonreactive(f) {
      var previous = Tracker.currentComputation;
      setCurrentComputation(null);
      try {
        return f();
      } finally {
        setCurrentComputation(previous);
      }
    }
  }, {
    key: "onInvalidate",
    value: function onInvalidate(f) {
      if (!Tracker.active) throw new Error("Tracker.onInvalidate requires a currentComputation");

      Tracker.currentComputation.onInvalidate(f);
    }
  }, {
    key: "afterFlush",
    value: function afterFlush(f) {
      afterFlushCallbacks.push(f);
      requireFlush();
    }
  }]);

  return Tracker;
}();

Tracker.active = false;
Tracker.currentComputation = null;
Tracker._computations = {};

Tracker.Computation = function () {
  function Computation(f, parent, onError) {
    _classCallCheck(this, Computation);

    if (!constructingComputation) throw new Error("Tracker.Computation constructor is private; use Tracker.autorun");
    constructingComputation = false;

    var self = this;

    // http://docs.meteor.com/#computation_stopped

    /**
    * @summary True if this computation has been stopped.
    * @locus Client
    * @memberOf Tracker.Computation
    * @instance
    * @name  stopped
    */
    self.stopped = false;
    // http://docs.meteor.com/#computation_invalidated

    /**
    * @summary True if this computation has been invalidated (and not yet rerun), or if it has been stopped.
    * @locus Client
    * @memberOf Tracker.Computation
    * @instance
    * @name  invalidated
    * @type {Boolean}
    */
    self.invalidated = false;
    // http://docs.meteor.com/#computation_firstrun

    /**
    * @summary True during the initial run of the computation at the time `Tracker.autorun` is called, and false on subsequent reruns and at other times.
    * @locus Client
    * @memberOf Tracker.Computation
    * @instance
    * @name  firstRun
    * @type {Boolean}
    */
    self.firstRun = true;

    self._id = nextId++;
    self._onInvalidateCallbacks = [];
    self._onStopCallbacks = [];
    // the plan is at some point to use the parent relation
    // to constrain the order that computations are processed
    self._parent = parent;
    self._func = f;
    self._onError = onError;
    self._recomputing = false;

    // Register the computation within the global Tracker.
    Tracker._computations[self._id] = self;

    var errored = true;
    try {
      self._compute();
      errored = false;
    } finally {
      self.firstRun = false;
      if (errored) self.stop();
    }
  }

  /**
   * @summary Registers `callback` to run when this computation is next invalidated, or runs it immediately if the computation is already invalidated.  The callback is run exactly once and not upon future invalidations unless `onInvalidate` is called again after the computation becomes valid again.
   * @locus Client
   * @param {Function} callback Function to be called on invalidation. Receives one argument, the computation that was invalidated.
   */


  _createClass(Computation, [{
    key: "onInvalidate",
    value: function onInvalidate(f) {
      var self = this;

      if (typeof f !== 'function') throw new Error("onInvalidate requires a function");

      if (self.invalidated) {
        Tracker.nonreactive(function () {
          withNoYieldsAllowed(f)(self);
        });
      } else {
        self._onInvalidateCallbacks.push(f);
      }
    }

    /**
     * @summary Registers `callback` to run when this computation is stopped, or runs it immediately if the computation is already stopped.  The callback is run after any `onInvalidate` callbacks.
     * @locus Client
     * @param {Function} callback Function to be called on stop. Receives one argument, the computation that was stopped.
     */

  }, {
    key: "onStop",
    value: function onStop(f) {
      var self = this;

      if (typeof f !== 'function') throw new Error("onStop requires a function");

      if (self.stopped) {
        Tracker.nonreactive(function () {
          withNoYieldsAllowed(f)(self);
        });
      } else {
        self._onStopCallbacks.push(f);
      }
    }

    /**
     * @summary Invalidates this computation so that it will be rerun.
     * @locus Client
     */

  }, {
    key: "invalidate",
    value: function invalidate() {
      var self = this;
      if (!self.invalidated) {
        // if we're currently in _recompute(), don't enqueue
        // ourselves, since we'll rerun immediately anyway.
        if (!self._recomputing && !self.stopped) {
          requireFlush();
          pendingComputations.push(this);
        }

        self.invalidated = true;

        // callbacks can't add callbacks, because
        // self.invalidated === true.
        for (var i = 0, f; f = self._onInvalidateCallbacks[i]; i++) {
          Tracker.nonreactive(function () {
            withNoYieldsAllowed(f)(self);
          });
        }
        self._onInvalidateCallbacks = [];
      }
    }
  }, {
    key: "stop",
    value: function stop() {
      var self = this;

      if (!self.stopped) {
        self.stopped = true;
        self.invalidate();
        // Unregister from global Tracker.
        delete Tracker._computations[self._id];

        var _loop = function _loop(i, f) {
          Tracker.nonreactive(function () {
            withNoYieldsAllowed(f)(self);
          });
        };

        for (var i = 0, f; f = self._onStopCallbacks[i]; i++) {
          _loop(i, f);
        }
        self._onStopCallbacks = [];
      }
    }
  }, {
    key: "_compute",
    value: function _compute() {
      var self = this;
      self.invalidated = false;

      var previous = Tracker.currentComputation;
      setCurrentComputation(self);
      var previousInCompute = inCompute;
      inCompute = true;
      try {
        withNoYieldsAllowed(self._func)(self);
      } finally {
        setCurrentComputation(previous);
        inCompute = previousInCompute;
      }
    }
  }, {
    key: "_needsRecompute",
    value: function _needsRecompute() {
      var self = this;
      return self.invalidated && !self.stopped;
    }
  }, {
    key: "_recompute",
    value: function _recompute() {
      var self = this;

      console.log(self);

      self._recomputing = true;
      try {
        if (self._needsRecompute()) {
          try {
            self._compute();
          } catch (e) {
            if (self._onError) {
              self._onError(e);
            } else {
              _throwOrLog("recompute", e);
            }
          }
        }
      } finally {
        self._recomputing = false;
      }
    }
  }]);

  return Computation;
}();

Tracker.Dependency = function Dependency() {
  _classCallCheck(this, Dependency);

  this.depend = function (computation) {
    if (!computation) {
      if (!Tracker.active) return false;

      computation = Tracker.currentComputation;
    }
    var self = this;
    var id = computation._id;
    if (!(id in self._dependentsById)) {
      self._dependentsById[id] = computation;
      computation.onInvalidate(function () {
        delete self._dependentsById[id];
      });
      return true;
    }
    return false;
  };

  this.changed = function () {
    var self = this;
    for (var id in self._dependentsById) {
      self._dependentsById[id].invalidate();
    }
  };

  this.hasDependents = function () {
    var self = this;
    for (var id in self._dependentsById) {
      return true;
    }return false;
  };

  this._dependentsById = {};
};

exports.default = Tracker;


var withNoYieldsAllowed = function withNoYieldsAllowed(f) {
  if (typeof Meteor === 'undefined' || Meteor.isClient) {
    return f;
  } else {
    return function () {
      var args = arguments;
      Meteor._noYieldsAllowed(function () {
        f.apply(null, args);
      });
    };
  }
};

var setCurrentComputation = function setCurrentComputation(c) {
  Tracker.currentComputation = c;
  Tracker.active = !!c;
};

var _debugFunc = function _debugFunc() {
  // We want this code to work without Meteor, and also without
  // "console" (which is technically non-standard and may be missing
  // on some browser we come across, like it was on IE 7).
  //
  // Lazy evaluation because `Meteor` does not exist right away.(??)
  return typeof Meteor !== "undefined" ? Meteor._debug : typeof console !== "undefined" && console.error ? function () {
    console.error.apply(console, arguments);
  } : function () {};
};

var _maybeSuppressMoreLogs = function _maybeSuppressMoreLogs(messagesLength) {
  // Sometimes when running tests, we intentionally suppress logs on expected
  // printed errors. Since the current implementation of _throwOrLog can log
  // multiple separate log messages, suppress all of them if at least one suppress
  // is expected as we still want them to count as one.
  if (typeof Meteor !== "undefined") {
    if (Meteor._suppressed_log_expected()) {
      Meteor._suppress_log(messagesLength - 1);
    }
  }
};

var _throwOrLog = function _throwOrLog(from, e) {
  if (throwFirstError) {
    throw e;
  } else {
    var printArgs = ["Exception from Tracker " + from + " function:"];
    if (e.stack && e.message && e.name) {
      var idx = e.stack.indexOf(e.message);
      if (idx < 0 || idx > e.name.length + 2) {
        // check for "Error: "
        // message is not part of the stack
        var message = e.name + ": " + e.message;
        printArgs.push(message);
      }
    }
    printArgs.push(e.stack);
    _maybeSuppressMoreLogs(printArgs.length);

    for (var i = 0; i < printArgs.length; i++) {
      _debugFunc()(printArgs[i]);
    }
  }
};

var nextId = 1;
// computations whose callbacks we should call at flush time
var pendingComputations = [];
// `true` if a Tracker.flush is scheduled, or if we are in Tracker.flush now
var willFlush = false;
// `true` if we are in Tracker.flush now
var inFlush = false;
// `true` if we are computing a computation now, either first time
// or recompute.  This matches Tracker.active unless we are inside
// Tracker.nonreactive, which nullfies currentComputation even though
// an enclosing computation may still be running.
var inCompute = false;
// `true` if the `_throwFirstError` option was passed in to the call
// to Tracker.flush that we are in. When set, throw rather than log the
// first error encountered while flushing. Before throwing the error,
// finish flushing (from a finally block), logging any subsequent
// errors.
var throwFirstError = false;

var afterFlushCallbacks = [];

// Tracker.Computation constructor is visible but private
// (throws an error if you try to call it)
var constructingComputation = false;

var requireFlush = function requireFlush() {
  if (!willFlush) {
    // We want this code to work without Meteor, see debugFunc above
    if (typeof Meteor !== "undefined") Meteor._setImmediate(Tracker._runFlush);else setTimeout(Tracker._runFlush, 0);
    willFlush = true;
  }
};